#include "Message.h"


void PathDigest::Compute( int num, in_addr * ip_list )
{
    digest1 = digest2 = 0;
    for ( int i = 0; i < num; i ++ )
    {
        if ( ip_list[i].s_addr != 0 )
        {
            digest1 += ip_list[i].s_addr;
            digest2 += ip_list[i].s_addr * ( i + 1 );
        }
    }
}


bool PathDigest::operator == ( const PathDigest& rhs ) const
{
    return digest1 == rhs.digest1 && digest2 == rhs.digest2;
}


void PathBitmap::SetPathBit( short bit_id )
{
    if ( bit_id < 1 || bit_id > ( PATH_BITMAP_SIZE << 3 ) )
        return;
    short byte_idx = ( bit_id - 1 ) >> 3;
    short bit_idx = ( bit_id - 1 ) & 0x7;
    char bit_editor = 0x1 << bit_idx;
    map[byte_idx] |= bit_editor;
}


void PathBitmap::ClearPathBit( short bit_id )
{
    if ( bit_id < 1 || bit_id > ( PATH_BITMAP_SIZE << 3 ) )
        return;
    short byte_idx = ( bit_id - 1 ) >> 3;
    short bit_idx = ( bit_id - 1 ) & 0x7;
    char bit_editor = 0x1 << bit_idx;
    map[byte_idx] &= ~bit_editor;
}


bool PathBitmap::IssetPathBit( short bit_id ) const
{
    if ( bit_id < 1 || bit_id > ( PATH_BITMAP_SIZE << 3 ) )
        return false;
    short byte_idx = ( bit_id - 1 ) >> 3;
    short bit_idx = ( bit_id - 1 ) & 0x7;
    char bit_editor = 0x1 << bit_idx;
    return ( ( map[byte_idx] & bit_editor ) != 0 );
}


void PathBitmap::Reset( void )
{
    bzero( map, PATH_BITMAP_SIZE );
}


bool CheckLniIntegrity( const LinkNodeInfo & lni )
{
    if ( lni.technique < TECH_NA || lni.technique > TECH_4P )
        return false;
    if ( lni.th_f.bin_index == -1 || lni.th_b.bin_index == -1 || lni.th_s.bin_index == -1 )
        return false;
    if ( ( lni.technique == TECH_4P || lni.technique == TECH_FSD ) && lni.th_d.bin_index == -1 )
        return false;
    return true;
}


bool CheckIntegrity( const Payload_MSG_LINKC * cong_event )
{
    if ( cong_event->total_hop_num <= 0 )
        return false;
    if ( cong_event->hop_num < 0 || cong_event->hop_num > cong_event->total_hop_num )
        return false;
    if ( cong_event->lni1.isset && ! CheckLniIntegrity( cong_event->lni1 ) )
        return false;
    if ( cong_event->lni2.isset && ! CheckLniIntegrity( cong_event->lni2 ) )
        return false;
    return true;
}
