#include "Pong_Daemon_Monitor.h"

char *TargetHost = NULL;
char *targetName;
struct in_addr targetAddress;
int sock_monitor;

int main( int argc, char **argv )
{
    ParseCommand( argc, argv );
    SetupMonitorClient();
    ProcessMonitorClient();

    return 0;
}


void Usage( char *argv0 )
{
    fprintf( stdout, "Usage: %s [host]\n", argv0 );
}


void ParseCommand( int argc, char **argv )
{
    int c;
    while ( ( c = getopt( argc, argv, "h" ) ) != -1 )
    {
	    switch ( c )
        {

       	// h: get help
	    case 'h':

	    // Didn't know how to handle this case
	    default:
            {
                Usage( argv[0] );
	            exit(1);
	        }
        }
    }
    if ( optind == (argc - 1) )
        TargetHost = argv[optind];
    else if ( optind == argc )
        TargetHost = "127.0.0.1";
    else
    {
	    Usage( argv[0] );
	    exit(1);
	}
}


void SetupMonitorClient( void )
{
    /* 
    Create monitor channel: TCP connection
    */

    if ( ResolveTargetName( TargetHost ) < 0 )
    {
	    exit(1);
    }

    if ( ( sock_monitor = socket( AF_INET, SOCK_STREAM, 0 ) ) < 0 )
    {
        perror( "socket( AF_INET, SOCK_STREAM )" );
        exit(1);
    }
    /*
    int opt_value = 1;
    if ( setsockopt( sock_monitor, SOL_SOCKET, SO_REUSEADDR, (char *)&opt_value, sizeof( opt_value ) ) < 0 )
    {
        perror( "setsockopt( sock_monitor, SOL_SOCKET, SO_REUSEADDR )" );
        exit(1);
    }
    */

    struct sockaddr_in monitor_addr;
    bzero( (char *)&monitor_addr, sizeof( monitor_addr ) );
    monitor_addr.sin_family = AF_INET;
    memcpy( (void *)&(monitor_addr.sin_addr.s_addr), &targetAddress, sizeof( targetAddress ) );
    monitor_addr.sin_port = htons( DMONITOR_PORT );
    if ( connect( sock_monitor, (struct sockaddr *)&monitor_addr, sizeof( monitor_addr ) ) < 0 )
    {
        perror( "Make sure that pong_daemon is running on this host" );
        exit(1);
    }
    
    fprintf( stderr, "Daemon control channel to '%s' established\n", targetName );
}


int ResolveTargetName( char *t )
{
    struct hostent *host;	// resolver host entry

    // Attempt to resolve, if possible
    host = gethostbyname( t );

    // Resolver failed
    if ( host == NULL )
    {

	    // Some systems don't have herror (non-BSD?), so for those,
	    // we'll cobble together an error message.
#ifdef HAVE_HERROR
	    herror( t );
#else
	    fprintf( stderr, "%s: Host not found\n", t );
#endif /* HAVE_HERROR */

	    memset( (void *)&targetAddress, 0, sizeof( struct in_addr ) );
	    targetName = strdup( t );

	    if ( targetName == NULL )
        {
	        fprintf( stderr, "Couldn't allocate memory for target hostname.\n" );
	        return -1;
	    }
	    return -1;
    }

    // Get IP address
    memcpy( &targetAddress, host->h_addr_list[0], host->h_length );

    // Make a copy of the canonical hostname
    targetName = strdup( host->h_name );
    if ( targetName == NULL )
    {
	    fprintf( stderr, "Couldn't allocate memory for target hostname.\n" );
	    return -1;
    }

    return 0;
}


void ProcessMonitorClient( void )
{
    int max_fd = MAX( sock_monitor, STDIN );
    while ( 1 )
    {
        fd_set read_fds;

        FD_ZERO( &read_fds );
        FD_SET( sock_monitor, &read_fds );
        FD_SET( STDIN, &read_fds );

        if ( select( max_fd + 1, &read_fds, NULL, NULL, NULL ) < 0 )
        {
	        perror( "select" );
	        exit(1);
	    }
        char buffer[2000];
        if ( FD_ISSET( sock_monitor, &read_fds ) )
        {
            int bytes = read( sock_monitor, buffer, 2000 );
            if ( bytes <= 0 )
                break;
            bool exit_client = false;
            if ( buffer[bytes - 1] == EOF )
            {
                bytes --;
                sock_monitor = -1;
                exit_client = true;
            }
            if ( bytes > 0 )
                write( STDOUT, buffer, bytes );
            if ( exit_client )
                break;
        }
        if ( FD_ISSET( STDIN, &read_fds ) )
        {
            fgets( buffer, 200, stdin );
            write( sock_monitor, buffer, strlen( buffer ) );
        }
    }
    close( sock_monitor );
}
