#include "Pong_Daemon_v2.h"
#include "TMon.h"
#include "Pong_Report.h"

bool done = false;
int DebugLevel = 3;
time_t start_time, current_time = 0, checkpoint_time;
const char * mmt_data_dir = ".";
int idle_duration = 180;    // Maximum idle duration in minutes
const char * xID;
int sock_monitor_listener;  // Socket used to establish monitor channel
int sock_monitor;           // Socket of monitor channel
int sock_control_listener;  // Socket used to establish control channel
int sock_control;           // Socket of control channel
pthread_t tid_control;
int role = FREE;
int min_probe_num = 4;
int min_hop = 6;
int state = ST_FREE;
int heartbeat = -1, dt = -1;
const char * shell_script = "run_pong.sh";
int run_count = 0;
int boot_failure_count = 0;
int access_deny_count = 0;
int last_failure_reason = -1;
int measurement_duration = 1000000;     // Measurement duration in minutes, 694 days
int mmt_end_time = 0;       // Time when measurement should end, 0 if it is not set yet
int mmt_seq_no = 0;
bool revive = false;
int peer_id = 0;

int main( int argc, char **argv )
{
    ParseCommand( argc, argv );
    CheckProcess();
    PrintPID();
    ReadHostID();
    start_time = checkpoint_time = time( NULL );
    InitializeArchive();
    signal( SIGTERM, Sig_Term );
    SetupMonitorServer();   // Set up TCP monitor channel
    SetupControlServer();   // Set up TCP control channel
    ReportInitialize();     // Set up UDP report channel
    SetupControlProcess();  // Start the thread that receive messages
    SetupTMonProcess();     // Start the TMon thread

    PrintXID( TRAN_START );
    RunDaemon();

    PrintXID( TRAN_END );
    IF_DEBUG( 1, fprintf( stderr, "Pong daemon has quitted.\n" ) );

    //pthread_join( tid_tmon, NULL );
    //pthread_join( tid_control, NULL );

    ReportFinalize();
    close( sock_monitor_listener );
    close( sock_control_listener );

    return 0;
}


void RunDaemon( void )
{
    int maximum_seconds = idle_duration * 60;
    time_t check_10sec = 0;

    while ( ! done )
    {
        sleep( 1 );
        current_time = time( NULL );

        // Check pending report messages every 1 second
        CheckPendingMessages( current_time );

        if ( current_time - check_10sec >= 10 )
        {
            check_10sec = current_time;

            // ReportTest();

            // *****************************************
            // Check the following part every 10 seconds
            // *****************************************

            if ( current_time - checkpoint_time > maximum_seconds )
            {
                done = true;
                break;
            }

            const char * mt = GetMeasurementTime();
            heartbeat = CheckPongHeartBeat( dt );

            int previous_state = state;

            switch( state )
            {
            case ST_SND_TRY:
            case ST_RCV_TRY:
            case ST_SND_RUN:
            case ST_RCV_RUN:
                {
                    if ( mmt_end_time && current_time > mmt_end_time )
                    {
                        IF_DEBUG( 3, fprintf( stderr, "%s Measurement time is up\n", mt ) );
                        TerminatePong();
                        if ( state == ST_SND_RUN || state == ST_RCV_RUN )
                            SaveMeasurementData( E_TIME_UP );
                        FreeState();
                    }
                    else
                    {
                        switch ( heartbeat )
                        {
                        case HB_RUN:
                            if ( state == ST_SND_TRY || state == ST_RCV_TRY )
                            {
                                if ( ! mmt_end_time )
                                {
                                    IF_DEBUG( 3, fprintf( stderr, "%s Measurement has started\n", mt ) );
                                    mmt_end_time = current_time + measurement_duration * 60;
                                }
                                boot_failure_count = 0;
                                access_deny_count = 0;
                                state = ( state == ST_SND_TRY ? ST_SND_RUN : ST_RCV_RUN );
                            }
                            else if ( state == ST_SND_RUN || state == ST_RCV_RUN )
                            {
                                bool pong_dead = false;
                                if ( ! PongAlive() )
                                {
                                    IF_DEBUG( 3, fprintf( stderr, "%s Pong is dead\n", mt ) );
                                    pong_dead = true;
                                }
                                else
                                {
                                    if ( dt < TH_DT_RUN )
                                    {
                                        IF_DEBUG( 9, fprintf( stderr, "%s Pong is running, dt=%d(sec)\n", mt, dt ) );
                                    }
                                    else
                                    {
                                        IF_DEBUG( 3, fprintf( stderr, "%s Pong has no heartbeat for %d(sec)\n", mt, dt ) );
                                        pong_dead = true;
                                        TerminatePong();
                                    }
                                }
                                if ( pong_dead )
                                {
                                    SaveMeasurementData( E_NO_HB );
                                    state = ST_TMP_FAIL;
                                }
                            }
                            break;

                        case HB_BOOT:
                            if ( state == ST_SND_TRY || state == ST_RCV_TRY )
                            {
                                bool pong_dead = false;
                                if ( ! PongAlive() )
                                {
                                    IF_DEBUG( 3, fprintf( stderr, "%s Pong is dead\n", mt ) );
                                    pong_dead = true;
                                }
                                else
                                {
                                    int th_dt_boot = ( state == ST_SND_TRY ? TH_DT_BOOT_SND : TH_DT_BOOT_RCV );
                                    if ( dt < th_dt_boot )
                                    {
                                        IF_DEBUG( 8, fprintf( stderr, "%s Pong is booting, dt=%d(sec)\n", mt, dt ) );
                                    }
                                    else
                                    {
                                        IF_DEBUG( 3, fprintf( stderr, "%s Pong is in booting state, but has no heartbeat for %d(sec)\n", mt, dt ) );
                                        pong_dead = true;
                                        TerminatePong();
                                    }
                                }
                                if ( pong_dead )
                                {
                                    if ( ++ boot_failure_count >= TH_BOOT_FAIL )
                                        state = ST_PMT_FAIL;
                                    else
                                        state = ST_TMP_FAIL;
                                }
                            }
                            break;

                        case HB_RMTOD:
                        case HB_CANCEL:
                        case HB_CNCL_RA:
                        case HB_CNCL_IHN:
                            state = ST_CANCEL;
                            break;
                        
                        case HB_ACDENY:
                            if ( ++ access_deny_count >= TH_ACCESS_DENY )
                                state = ST_PMT_FAIL;
                            else
                            {
                                sleep( 20 );
                                state = ST_TMP_FAIL;
                            }
                            break;

                        case HB_TRA:
                            state = ST_TRA;
                            break;

                        case HB_CORRUPT:
                        case HB_NOCONN:
                            if ( ++ boot_failure_count >= TH_BOOT_FAIL )
                                state = ST_PMT_FAIL;
                            else
                            {
                                sleep( 10 );
                                state = ST_TMP_FAIL;
                            }
                            break;

                        case HB_QUITTED:
                            if ( state == ST_SND_RUN || state == ST_RCV_RUN )
                                SaveMeasurementData( E_QUIT );
                            state = ST_TMP_FAIL;
                            break;

                        case HB_QUIT_RA:
                            if ( state == ST_SND_RUN || state == ST_RCV_RUN )
                                SaveMeasurementData( E_RA );
                            state = ST_TMP_FAIL;
                            break;

                        case HB_SIG_TERM:
                            if ( state == ST_SND_RUN || state == ST_RCV_RUN )
                                SaveMeasurementData( E_SIG_TERM );
                            state = ST_FREE;
                            break;

                        case HB_OUTDATE:
                            state = ST_OUTDATE;
                            IF_DEBUG( 1, fprintf( stderr, "Pong binary on this machine is outdated\n" ) );
                            break;
                        }
                    }
                }
                break;

            case ST_TRA:
                {
                    if ( current_time > mmt_end_time )
                    {
                        assert( mmt_end_time );
                        IF_DEBUG( 3, fprintf( stderr, "%s Measurement time is up\n", mt ) );
                        TerminatePong();
                        if ( role == SENDER || role == RECEIVER )
                            SaveMeasurementData( E_TIME_UP );
                        FreeState();
                    }
                    else
                    {
                        switch ( heartbeat )
                        {
                        case HB_TRA:
                            if ( dt < TH_DT_RUN )
                            {
                                IF_DEBUG( 9, fprintf( stderr, "%s Pong is tracing RA, dt=%d(sec)\n", mt, dt ) );
                            }
                            else
                            {
                                IF_DEBUG( 3, fprintf( stderr, "%s Pong has no heartbeat for %d(sec)\n", mt, dt ) );
                                if ( role == SENDER || role == RECEIVER )
                                    SaveMeasurementData( E_RA_NO_HB );
                                state = ST_TMP_FAIL;
                            }
                            break;

                        case HB_SIG_TERM:
                            if ( role == SENDER || role == RECEIVER )
                                SaveMeasurementData( E_SIG_TERM );
                            state = ST_FREE;
                            break;

                        case HB_CORRUPT:
                            if ( ++ boot_failure_count >= TH_BOOT_FAIL )
                                state = ST_PMT_FAIL;
                            else
                                state = ST_TMP_FAIL;
                            break;

                        case HB_QUIT_RA:
                            if ( role == SENDER || role == RECEIVER )
                                SaveMeasurementData( E_RA );
                            state = ST_TMP_FAIL;
                            break;

                        default:
                            IF_DEBUG( 1, fprintf( stderr, "%s Unexpected heartbeat type (%d) for state ST_TRA\n", mt, heartbeat ) );
                        }
                    }
                }
                break;
            }

            if ( state == ST_TMP_FAIL && ( role == SENDER || role == RECEIVER ) )
            {
                last_failure_reason = heartbeat;
                IF_DEBUG( 3, fprintf( stderr, "%s Restart running Pong, failure_reason = %d\n", mt, last_failure_reason ) );
                RunPong();
                if ( role == SENDER )
                    state = ST_SND_TRY;
                else
                    state = ST_RCV_TRY;
            }
            else if ( state == ST_PMT_FAIL && previous_state != ST_PMT_FAIL )
            {
                last_failure_reason = heartbeat;
                IF_DEBUG( 3, fprintf( stderr, "%s Permanent failure encounterd, failure_reason = %d\n", mt, last_failure_reason ) );
                // Send MSG_PONG_FAIL
                tmon->SendPongFailure( role, peer_id, state );
            }
            else if ( state == ST_CANCEL && previous_state != ST_CANCEL )
            {
                last_failure_reason = heartbeat;
                IF_DEBUG( 3, fprintf( stderr, "%s Cancel measurement, failure_reason = %d\n", mt, last_failure_reason ) );
                // Send MSG_PONG_FAIL
                tmon->SendPongFailure( role, peer_id, state );
            }
        }
    }

    TerminatePong();
    if ( state == ST_SND_RUN || state == ST_RCV_RUN )
        SaveMeasurementData( E_DMN_TERM );
}


void TerminatePong( void )
{
    pid_t pid = PongAlive();
    if ( pid )
    {
        kill( pid, SIGTERM );
        sleep( 3 );
        if ( PongAlive( pid ) )
            kill( pid, SIGKILL );
    }
}


void Usage( char *argv0 )
{
    fprintf( stdout, "Usage: %s [-d debuglevel] [-m dir] [-t min] [-T min] [-r] [-l host] [-i interface] [-H host_list] xID\n", argv0 );
    fprintf( stdout, "  -d debuglevel: set initial debugging level\n" );
    fprintf( stdout, "         -m dir: set directory of measurement data, default is '.'\n" );
    fprintf( stdout, "         -t min: specify the maximum idle duration in minutes, default is 1440, i.e., one day\n" );
    fprintf( stdout, "         -T min: specify the measurement duration in minutes, default is 100\n" );
    fprintf( stdout, "             -r: revival mode, don't clean up archives\n" );
    
    // *******************************************
    // Parameters for TMon
    fprintf( stdout, "        -l host: specify local ip/host\n" );
    fprintf( stdout, "   -i interface: specify local interface\n" );
    fprintf( stdout, "  -H  host_list: filename of host list\n" );
    // *******************************************

    fprintf( stdout, "            xID: transaction ID\n" );
}


void ParseCommand( int argc, char **argv )
{
    int c;
    while ( ( c = getopt( argc, argv, "hd:m:t:T:rl:i:p:H:" ) ) != -1 )
    {
	    switch ( c )
        {
	    case 'd':
            {
	            DebugLevel = atoi( optarg );
	            break;
	        }

        case 'm':
            {
                mmt_data_dir = strdup( optarg );
                break;
            }

        case 't':
            {
	            idle_duration = atoi( optarg );
	            break;
	        }

        case 'T':
            {
	            measurement_duration = atoi( optarg );
	            break;
	        }

        case 'r':
            {
                revive = true;
                break;
            }

        // l: set local origin of probes
        case 'l':
            {
                OriginHost = strdup( optarg );
                break;
            }
            
        case 'i':
            {
#ifdef HAVE_PCAP
                Interface = strdup( optarg );
                if ( Interface == NULL )
                {
                    fprintf(stderr, "Couldn't allocate space for interface name\n");
                    assert(0);
                }
                break;
#else
                fprintf( stderr, "pcap unavailable in this build" );
                exit(1);
#endif // HAVE_PCAP
            }

        case 'H':
            {
                TargetHostList = strdup( optarg );
                break;
            }

       	// h: Get help
	    case 'h':

	    // Didn't know how to handle this case
	    default:
            {
                Usage( argv[0] );
	            exit(1);
	        }
        }
    }
    if ( optind != (argc - 1) )
    {
	    Usage( argv[0] );
	    exit(1);
	}
    xID = argv[optind];
}


void HelpInteractiveCommand( void )
{
    fprintf( stdout, "***** Interactive commands helper *****\n" );
    fprintf( stdout, "x          - Exit control client\n" );
    fprintf( stdout, "q          - Query\n" );
    fprintf( stdout, "quit       - Quit\n" );
    fprintf( stdout, "h          - Help\n" );
    fprintf( stdout, "s          - Run 'pong_snd'\n" );
    fprintf( stdout, "r          - Run 'pong_rcv'\n" );
    fprintf( stdout, "e          - End 'pong_snd' or 'pong_rcv'\n" );
    fprintf( stdout, "***************************************\n" );
    fflush( stdout );
}


void ProcessInteractiveCommand( void )
{
    char input[250];
    bool valid;
    if( fgets( input, 250, stdin ) == NULL )
    {
        sock_monitor = -1;
        return;
    }
    int input_len = strlen( input ) - 1;
    switch ( input[0] )
    {
    case 'q':
        if ( input_len == 1 )
            Query();
        else if ( strcmp( input, "quit\n" ) == 0 )
            Quit();
        break;

    case 'x':
        TearDownClient();
        break;

    case 'h':
        HelpInteractiveCommand();
        break;

    case 's':
        char receiver_id[10];
        valid = false;
        if ( input_len > 2 )
            valid = ( sscanf( input + 2, "%s %d", receiver_id, &mmt_seq_no ) == 2 );
        if ( ! valid )
        {
            fprintf( stdout, "'s' command: Run 'pong_snd'\n" );
            fprintf( stdout, "     Syntax: s <receiver_id> <seq_no>\n" );
            fflush( stdout );
        }
        else
        {
            RunSender( receiver_id );
        }
        break;

    case 'r':
        char target_host[200];
        valid = false;
        if ( input_len > 2 )
            valid = ( sscanf( input + 2, "%s %d", target_host, &mmt_seq_no ) == 2 );
        if ( ! valid )
        {
            fprintf( stdout, "'r' command: Run 'pong_rcv'\n" );
            fprintf( stdout, "     Syntax: r <target_host> <seq_no>\n" );
            fflush( stdout );
        }
        else
        {
            RunReceiver( target_host );
        }
        break;

    case 'e':
        EndPong();
        break;
    }
}


void ProcessBinaryCommand( void )
{
    short ctr_code = RecvCtrCode();
    switch ( ctr_code )
    {
    case MCC_END:
        fprintf( stderr, "MCC_END received\n" );
        break;
    }
}


//********************************************
// Sig_Term( sig_num )
// -------------------------------------------
// The signal handler for SIGTERM
//********************************************

void Sig_Term( int sig_num )
{
    done = true;
}


void InitializeArchive( void )
{
    system( "./kpong.sh" );
    if ( revive )
    {
        char time_string[50];
        strcpy( time_string, asctime( gmtime( &start_time ) ) );
        time_string[strlen(time_string) - 1] = 0;
        IF_DEBUG( 3, fprintf( stderr, "Daemon process is revived at %s UTC.\n", time_string ) );
    }
    else
    {
        system( "if [ \"$(ls -A history)\" ]; then rm -r history/*; fi" );
        remove( "data/summary" );
        remove( "data/summary_rcv" );
        remove( "route_alteration" );
    }
    remove( "pong_pid" );
    remove( "heartbeat" );
    remove( "pong_data.tgz" );
}


void PrintPID( void )
{
    const char * file = "pong_daemon_pid";
    FILE * fp;
    if ( ( fp = fopen( file, "w" ) ) == NULL )
    {
        fprintf( stderr, "Fail to write file '%s'\n", file );
        exit(1);
    }

    pid_t pid = getpid();
    fprintf( fp, "%d", pid );
    fclose( fp );
}


void PrintXID( int transaction_type )
{
    const char * file = "xID";
    FILE * fp;
    if ( ( fp = fopen( file, "w" ) ) == NULL )
        fprintf( stderr, "Fail to write file '%s'\n", file );

    const char * tran_string = "Unknown Operation";
    
    switch ( transaction_type )
    {
    case TRAN_START:
        tran_string = "Run Daemon";
        break;
    case TRAN_END:
        tran_string = "End Pong";
        break;
    }

    fprintf( fp, "%s\t%s\n", xID, tran_string );
    fclose( fp );
}


void CheckProcess( void )
{
    const char * file = "pong_daemon_pid";
    FILE * fp;
    if ( ( fp = fopen( file, "r" ) ) != NULL )
    {
        pid_t pid;
        fscanf( fp, "%d", &pid );
        fclose( fp );
        if ( kill( pid, 0 ) == 0 )
        {
            fprintf( stderr, "Pong daemon is already running\n" );
            exit(1);
        }
    }
}


// Check whether pong_snd or pong_rcv is alive or not

pid_t PongAlive( void )
{
    const char * file = "pong_pid";
    FILE * fp;
    if ( ( fp = fopen( file, "r" ) ) != NULL )
    {
        pid_t pid;
        fscanf( fp, "%d", &pid );
        fclose( fp );
        if ( kill( pid, 0 ) == 0 )
            return pid;
    }
    return 0;
}


bool PongAlive( pid_t pid )
{
    if ( kill( pid, 0 ) == 0 )
        return true;
    return false;
}


// Check and read data from file 'heartbeat'

int CheckPongHeartBeat( int& dt )
{
    const char * file = "heartbeat";
    FILE * fp;
    if ( ( fp = fopen( file, "r" ) ) == NULL )
        return HB_QUITTED;

    char status[20], pong_type[20], comment[50];
    int return_code = HB_CORRUPT;
    int tv_sec;
    dt = -1;
    if ( fscanf( fp, "%s\t%d\t%s\t%s\n", status, &tv_sec, pong_type, comment ) == 4 )
    {
        dt = current_time - tv_sec;
        if ( strcmp( status, "booting" ) == 0 )
            return_code = HB_BOOT;
        else if ( strcmp( status, "running" ) == 0 )
            return_code = HB_RUN;
        else if ( strcmp( status, "quitted" ) == 0 )
        {
            if ( strcmp( comment, "route_alteration" ) == 0 )
                return_code = HB_QUIT_RA;
            else if ( strcmp( comment, "sigterm" ) == 0 )
                return_code = HB_SIG_TERM;
            else
                return_code = HB_QUITTED;
        }
        else if ( strcmp( status, "cancelled" ) == 0 )
        {
            if ( strcmp( comment, "route_alteration" ) == 0 )
                return_code = HB_CNCL_RA;
            else if ( strcmp( comment, "invalid_hop_num" ) == 0 )
                return_code = HB_CNCL_IHN;
            else
                return_code = HB_CANCEL;
        }
        else if ( strcmp( status, "no_conn" ) == 0 )
            return_code = HB_NOCONN;
        else if ( strcmp( status, "trace_ra" ) == 0 )
            return_code = HB_TRA;
        else if ( strcmp( status, "access_deny" ) == 0 )
            return_code = HB_ACDENY;
        else if ( strcmp( status, "peer_outdated" ) == 0 )
            return_code = HB_RMTOD;
        else if ( strcmp( status, "outdated" ) == 0 )
            return_code = HB_OUTDATE;

        if ( strcmp( pong_type, "sender" ) != 0 && strcmp( pong_type, "receiver" ) != 0 )
            return_code = HB_CORRUPT;
    }

    fclose( fp );
    return return_code;
}


void SetupControlProcess( void )
{
    if ( pthread_create( &tid_control, NULL, control_process, NULL ) != 0 )
    {
        fprintf( stderr, "pthread_create(control_process) failed\n" );
        exit(1);
    }
    IF_DEBUG( 3, fprintf( stderr, "Control process installed\n" ) );
}


void * control_process( void * arg )
{
    ReceiveCommands();

    pthread_exit(NULL);
    
    return NULL;
}


// Receive interactive commands

void ReceiveCommands( void )
{
    int init_max_fd = MAX( sock_monitor_listener, sock_control_listener );
    init_max_fd = MAX( init_max_fd, sock_report );

    while ( ! done )
    {
        int max_fd = init_max_fd;
        fd_set read_fds;

        FD_ZERO( &read_fds );
        FD_SET( sock_monitor_listener, &read_fds );
        FD_SET( sock_control_listener, &read_fds );
        FD_SET( sock_report, &read_fds );
        if ( sock_monitor >= 0 )
            FD_SET( STDIN, &read_fds );
        if ( sock_control >= 0 )
        {
            FD_SET( sock_control, &read_fds );
            max_fd = MAX( sock_control, max_fd );
        }

        int rc;
        rc = select( max_fd + 1, &read_fds, NULL, NULL, NULL );
        if ( rc < 0 )
        {
	        perror( "select" );
	        exit(1);
        }
        
        current_time = time( NULL );

        if ( FD_ISSET( STDIN, &read_fds ) )
        {
            ProcessInteractiveCommand();
        }
        if ( FD_ISSET( sock_monitor_listener, &read_fds ) )
        {
            AcceptMonitorClient();
        }
        if ( FD_ISSET( sock_control_listener, &read_fds ) )
        {
            AcceptControlClient();
        }
        if ( sock_control >= 0 && FD_ISSET( sock_control, &read_fds ) )
        {
            ProcessBinaryCommand();
        }
        if ( FD_ISSET( sock_report, &read_fds ) )
        {
            RecvUdpMessages();
        }
    }
}


void SetupMonitorServer( void )
{
    /* Control stream: TCP connection */
    if ( ( sock_monitor_listener = socket( AF_INET, SOCK_STREAM, 0 ) ) < 0 )
    {
        perror( "socket( AF_INET, SOCK_STREAM, 0 )" );
        exit(1);
    }
    int opt_value = 1;
    if ( setsockopt( sock_monitor_listener, SOL_SOCKET, SO_REUSEADDR, (char *)&opt_value, sizeof( opt_value ) ) < 0 )
    {
        perror( "setsockopt( SOL_SOCKET,SO_REUSEADDR )" );
        exit(1);
    }
    struct sockaddr_in monitor_addr;
    bzero( (char *)&monitor_addr, sizeof( monitor_addr ) );
    monitor_addr.sin_family = AF_INET;
    monitor_addr.sin_addr.s_addr = htonl( INADDR_ANY );
    monitor_addr.sin_port = htons( DMONITOR_PORT );
    if ( bind( sock_monitor_listener, (struct sockaddr *)&monitor_addr, sizeof( monitor_addr ) ) < 0 )
    {
        perror( "bind( sock_monitor_listener )" );
        exit(1);
    }
    if ( listen( sock_monitor_listener, 1 ) < 0 )
    {
        perror( "listen( sock_monitor_listener, 1 )" );
        exit(1);
    }
    sock_monitor = -1;
}


void SetupControlServer( void )
{
    /* Control stream: TCP connection */
    if ( ( sock_control_listener = socket( AF_INET, SOCK_STREAM, 0 ) ) < 0 )
    {
        perror( "socket( AF_INET, SOCK_STREAM, 0 )" );
        exit(1);
    }
    int opt_value = 1;
    if ( setsockopt( sock_control_listener, SOL_SOCKET, SO_REUSEADDR, (char *)&opt_value, sizeof( opt_value ) ) < 0 )
    {
        perror( "setsockopt( SOL_SOCKET,SO_REUSEADDR )" );
        exit(1);
    }
    struct sockaddr_in control_addr;
    bzero( (char *)&control_addr, sizeof( control_addr ) );
    control_addr.sin_family = AF_INET;
    control_addr.sin_addr.s_addr = htonl( INADDR_ANY );
    control_addr.sin_port = htons( DMONCNTL_PORT );
    if ( bind( sock_control_listener, (struct sockaddr *)&control_addr, sizeof( control_addr ) ) < 0 )
    {
        perror( "bind( sock_control_listener )" );
        exit(1);
    }
    if ( listen( sock_control_listener, 1 ) < 0 )
    {
        perror( "listen( sock_control_listener, 1 )" );
        exit(1);
    }
    sock_control = -1;
}


void AcceptMonitorClient( void )
{
    sock_monitor = accept( sock_monitor_listener, NULL, NULL );
    if ( sock_monitor < 0 )
    {
        perror( "accept( sock_monitor_listener )" );
        return;
    }
    fclose( stdin );
    if ( dup2( sock_monitor, STDIN ) < 0 )
    {
        perror( "dup2( STDIN )" );
        close( sock_monitor );
        sock_monitor = -1;
        return;
    }
    stdin = fdopen( STDIN, "r" );
    if ( dup2( sock_monitor, STDOUT ) < 0 )
    {
        perror( "dup2( STDOUT )" );
        close( sock_monitor );
        sock_monitor = -1;
        return;
    }
    close( sock_monitor );
}


void AcceptControlClient( void )
{
    if ( sock_control >= 0 )
        // This should never happen
        close( sock_control );

    sock_control = accept( sock_control_listener, NULL, NULL );
    if ( sock_control < 0 )
    {
        perror( "accept( sock_control_listener )" );
        return;
    }
}


void RunSender( const char * receiver_id, bool silent )
{
    if ( ! silent )
    {
        fprintf( stdout, "PONGC[RUN_SENDER]\n" );
        fflush( stdout );
    }

    FreeState();
    role = SENDER;
    state = ST_SND_TRY;

    IF_DEBUG( 3, fprintf( stderr, "%s Start running 'pong_snd' (receiver_id=%s seq_no=%d) ... ", GetMeasurementTime(), receiver_id, mmt_seq_no ) );
    GenerateShellScript( SENDER, receiver_id );
    RunPong();
    IF_DEBUG( 3, fprintf( stderr, "done.\n" ) );
}


void RunReceiver( const char * target_host, bool silent )
{
    if ( ! silent )
    {
        fprintf( stdout, "PONGC[RUN_RECEIVER]\n" );
        fflush( stdout );
    }

    FreeState();
    role = RECEIVER;
    state = ST_RCV_TRY;

    IF_DEBUG( 3, fprintf( stderr, "%s Start running 'pong_rcv' (target=%s, seq_no=%d) ... ", GetMeasurementTime(), target_host, mmt_seq_no ) );
    GenerateShellScript( RECEIVER, target_host );
    RunPong();
    IF_DEBUG( 3, fprintf( stderr, "done.\n" ) );
}


void EndPong( bool silent )
{
    if ( ! silent )
    {
        fprintf( stdout, "PONGC[END_PONG]\n" );
        fflush( stdout );
    }

    if ( role == FREE )
        return;
    const char * pong_program = ( role == SENDER ? "pong_snd" : "pong_rcv" );
    IF_DEBUG( 3, fprintf( stderr, "%s End '%s' ... ", GetMeasurementTime(), pong_program ) );
    TerminatePong();
    if ( state == ST_SND_RUN || state == ST_RCV_RUN )
        SaveMeasurementData( E_SIG_TERM );
    IF_DEBUG( 3, fprintf( stderr, "done.\n" ) );
    FreeState();
}


void FreeState( void )
{
    state = ST_FREE;
    role = FREE;
    mmt_end_time = 0;
    boot_failure_count = 0;
    access_deny_count = 0;
}


void Query( void )
{
    fprintf( stdout, "PONGC[QUERY]\t%s\t%ld\t%d\t%d\t%d\t%d\t%d\t%d\t%d\n", xID, current_time, role, state, mmt_seq_no, last_failure_reason, dt, boot_failure_count, run_count );
    fflush( stdout );
    // Renew the checkpoint time
    checkpoint_time = current_time;
}


void TearDownClient( void )
{
    fprintf( stdout, "PONGC[EXIT]\n" );
    fputc( EOF, stdout );
    fflush( stdout );
    sock_monitor = -1;
}


void Quit( void )
{
    IF_DEBUG( 1, fprintf( stderr, "%s Quitting ... \n", GetMeasurementTime() ) );
    kill( 0, SIGTERM );
}


const char * GetMeasurementTime( void )
{
    static char mt[20];
    int measurement_time = ( current_time == 0 ? 0: current_time - start_time );
    sprintf( mt, "%d:%02d:%02d", measurement_time / 3600, measurement_time % 3600 / 60, measurement_time % 60 );
    return mt;
}


void GenerateShellScript( int the_role, const char * target_host )
{
    FILE * fp;
    if ( ( fp = fopen( shell_script, "w" ) ) == NULL )
    {
        fprintf( stderr, "Fail to write file '%s'\n", shell_script );
        exit(1);
    }
    fprintf( fp, "#This is the shell command used by pong_daemon to run Pong process\n" );
    if ( the_role == SENDER )
        fprintf( fp, "./pong_snd -d 3 -i eth0 -g -s -n -p %d -b %d -u %d -r %s 2>pong_snd_log &\n", TCPSND_PORT, min_probe_num, min_hop, target_host );
    else
        fprintf( fp, "./pong_rcv -d 3 -i eth0 -n -p %d -b %d -u %d %s 2>pong_rcv_log &\n", TCPSND_PORT, min_probe_num, min_hop, target_host );
    fclose( fp );
    chmod( shell_script, S_IRUSR | S_IWUSR | S_IXUSR );
}


void RunPong( void )
{
    system( "./kpong.sh" );
    remove( "heartbeat" );
    remove( "pong_pid" );

    char cmd[50];
    sprintf( cmd, "./%s", shell_script );
    system( cmd );
}


int SaveMeasurementData( int end_reason )
{
    assert ( role == SENDER || role == RECEIVER );

    run_count ++;

    IF_DEBUG( 3, fprintf( stderr, "%s Saving measurement data.\n", GetMeasurementTime() ) );

    char path[200];
    FILE * fp;

    strcpy( path, mmt_data_dir );
    strcat( path, "/data/data_id" );
    if ( ( fp = fopen( path, "r" ) ) == NULL )
        return -1;
    char buffer[100];
    fgets( buffer, 100, fp );
    int data_id;
    int rc = sscanf( buffer, "%d", &data_id );
    if ( rc != 1 )
    {
        fclose( fp );
        return -2;
    }
    
    fgets( buffer, 100, fp );
    char path_id[20];
    rc = sscanf( buffer, "%s", path_id );
    fclose( fp );
    if ( rc != 1 )
        return -5;

    if ( role == SENDER )
    {
        strcpy( path, mmt_data_dir );
        strcat( path, "/data/summary" );
        if ( ( fp = fopen( path, "a" ) ) == NULL )
            return -3;
        fprintf( fp, "%d\t%s\t%d\n", data_id, path_id, end_reason );
        fclose( fp );

        if ( ( fp = fopen( "save_data.sh", "w" ) ) == NULL )
            return -4;
        fprintf( fp, "mv pong_snd_log heartbeat %s/data/\n", mmt_data_dir );
        fprintf( fp, "mv %s/data/summary .\n", mmt_data_dir );
        fprintf( fp, "mv %s/data %s/history/%d\n", mmt_data_dir, mmt_data_dir, data_id );
        fprintf( fp, "mkdir %s/data\n", mmt_data_dir );
        fprintf( fp, "mv summary %s/data/\n", mmt_data_dir );
        fclose( fp );
    }
    else
    {
        strcpy( path, mmt_data_dir );
        strcat( path, "/data/summary_rcv" );
        if ( ( fp = fopen( path, "a" ) ) == NULL )
            return -3;
        fprintf( fp, "%d\t%s\t%d\n", data_id, path_id, end_reason );
        fclose( fp );

        if ( ( fp = fopen( "save_data.sh", "w" ) ) == NULL )
            return -4;
        fprintf( fp, "mv pong_rcv_log heartbeat %s/data/\n", mmt_data_dir );
        fprintf( fp, "mv %s/data/summary_rcv .\n", mmt_data_dir );
        fprintf( fp, "mv %s/data %s/history/%d\n", mmt_data_dir, mmt_data_dir, data_id );
        fprintf( fp, "mkdir %s/data\n", mmt_data_dir );
        fprintf( fp, "mv summary_rcv %s/data/\n", mmt_data_dir );
        fclose( fp );
    }

    chmod( "save_data.sh", S_IRUSR | S_IWUSR | S_IXUSR );
    system( "./save_data.sh" );

    return data_id;
}


long RecvCtrCode( void )
{
    long ctr_code;
    int rc = recv( sock_control, &ctr_code, sizeof(long), MSG_WAITALL );
    if ( rc == 0 )
    {
        // Remote side has closed the connection
        close( sock_control );
        sock_control = -1;
        return MCC_CLOSE;
    }
    if ( rc != sizeof(long) )
    {
        IF_DEBUG( 3, fprintf( stderr, "%s RecvCtrCode() failed, rc=%d\n", GetMeasurementTime(), rc ) );
        return -1;
    }
    return ntohl( ctr_code );
}


bool IsPongRunning( void )
{
    if ( state == ST_FREE || state == ST_CANCEL || state == ST_OUTDATE || state == ST_PMT_FAIL )
        return false;
    else
        return true;
}
