#include "Pong_Report.h"

int sock_report;
pthread_mutex_t sr_mutex;   // Mutex for sock_report
list<PendingMessage> pending_messages;
pthread_mutex_t pmg_mutex;  // Mutex for pending_messages
long message_id;
pthread_mutex_t mid_mutex;      // Mutex for 'message_id'

#ifdef TEST_ALONE

int DebugLevel = 9;
pthread_t tid_mh;
bool done = false;
time_t time_baseline = 0;
time_t current_time;

int main()
{
    SetReportSocket();
    SetupMessageHandler();
    for ( int i = 0; i < 10; i ++ )
    {
        char * message = CreateMessage( 1000 + i, 10, i, 16 );
        if ( message )
        {
            strncpy( message + sizeof( MsgHeader ), "1234567890123456", 16 );
            SendMessage( message );
        }
    }
    while( ! pending_messages.empty() )
    {
        current_time = time( NULL );
        CheckPendingMessages( current_time );
        sleep( 1 );
    }
    close( sock_report );
}


void ReceiveMessages( void )
{
    int max_fd = sock_report;

    while ( ! done )
    {
        fd_set read_fds;

        FD_ZERO( &read_fds );
        FD_SET( sock_report, &read_fds );

        int rc;
        rc = select( max_fd + 1, &read_fds, NULL, NULL, NULL );
        if ( rc < 0 )
        {
            perror( "select" );
            exit(1);
        }

        if ( FD_ISSET( sock_report, &read_fds ) )
        {
            RecvUdpMessages();
        }
   }
}


void SetupMessageHandler( void )
{
    if ( pthread_create( &tid_mh, NULL, message_handler, NULL ) != 0 )
    {
        fprintf( stderr, "pthread_create(msg_handler) failed\n" );
        exit(1);
    }
    IF_DEBUG( 3, fprintf( stderr, "Message handler installed\n" ) );
}


void * message_handler( void * arg )
{
    ReceiveMessages();

    pthread_exit(NULL);
    
    return NULL;
}


const char * GetMeasurementTime( void )
{
    static char mt[20];
    int measurement_time = ( time_baseline == 0 ? 0: current_time - time_baseline );
    sprintf( mt, "%d:%02d:%02d", measurement_time / 3600, measurement_time % 3600 / 60, measurement_time % 60 );
    return mt;
}

#endif


// Set up the UDP socket used to send report messages and receive control messages

void SetReportSocket( void )
{
    // Specify the domain name of your centralized control node here (for TPong)
    const char * t = "localhost";

    hostent *host;      // resolver host entry

    // Attempt to resolve
    host = gethostbyname( t );

    // Resolver failed
    if ( host == NULL )
    {

        // Some systems don't have herror (non-BSD?), so for those,
        // we'll cobble together an error message.
#ifdef HAVE_HERROR
        herror( t );
#else
        fprintf( stderr, "%s: Host not found\n", t );
#endif /* HAVE_HERROR */

        exit(1);
    }

    IF_DEBUG( 3, fprintf( stderr, "targetName = %s\n", host->h_name ) );
    IF_DEBUG( 3, fprintf( stderr, "targetAddress = %s\n", inet_ntoa( *((in_addr *)(host->h_addr_list[0])) ) ) );

    sockaddr_in target_udp_addr;
    memset( &target_udp_addr, 0, sizeof( sockaddr_in ) );
    // Note:  Only BSD4.3Reno and later have sin_len in struct
    // sockaddr_in, so we need to test for it.
#ifdef HAVE_SOCKADDR_SA_LEN
    target_udp_addr.sin_len = sizeof( sockaddr_in );
#endif /* HAVE_SOCKADDR_SA_LEN */
    target_udp_addr.sin_family = AF_INET;
    target_udp_addr.sin_port = htons( REPORT_PORT );
    memcpy( &target_udp_addr.sin_addr, host->h_addr_list[0], host->h_length );

    // Create UDP socket
    if ( ( sock_report = socket( AF_INET, SOCK_DGRAM, 0 ) ) < 0 )
    {
        perror( "socket( AF_INET, SOCK_DGRAM, 0 )" );
        exit(1);
    }

#ifdef linux
    int bsdcompatOption;
    bsdcompatOption = 1;
    if ( setsockopt( sock_report, SOL_SOCKET, SO_BSDCOMPAT, &bsdcompatOption, sizeof( bsdcompatOption ) ) < 0 )
    {
        perror( "setsockopt( SO_BSDCOMPAT )" );
        exit(1);
    }
#endif /* linux */

    // Bind remote side of socket
    if ( connect( sock_report, (sockaddr *)&target_udp_addr, sizeof( sockaddr_in ) ) < 0 )
    {
        perror( "connect" );
        exit(1);
    }
}


// Receive UDP messages from control node

void RecvUdpMessages( void )
{
    sockaddr_in remote_addr;
    socklen_t fromlen = sizeof( sockaddr_in );
    char buffer[MAX_MESSAGE_SIZE];
    MsgHeader * msg_header = (MsgHeader * )buffer;
    //char * payload = buffer + sizeof( MsgHeader );
    int msg_len = recvfrom( sock_report, buffer, MAX_MESSAGE_SIZE, 0, (sockaddr *)&remote_addr, &fromlen );
    if ( msg_len < 0 )
    {
        perror( "recvfrom" );
        return;
    }
    if ( (unsigned int)msg_len < sizeof( MsgHeader ) || msg_header->len + sizeof( MsgHeader ) != (unsigned int)msg_len )
    {
        fprintf( stderr, "%s RecvUdpMessages() failed, msg_len=%d\n", GetMeasurementTime(), msg_len );
        return;
    }
    if ( msg_header->signature != msg_signature )
    {
        fprintf( stderr, "%s RecvUdpMessages() failed, invalid signature\n", GetMeasurementTime() );
        return;
    }

    // Debug
    //fprintf( stderr, "Message received: nID=%d, type=%d, id=%ld, flag=%x, msg_len=%d, payload_len=%d\n", msg_header->nID, msg_header->type, msg_header->id, msg_header->flag, msg_len, msg_header->len );

    // Process the UDP message
    ProcessUdpMessage( buffer );

    if ( msg_header->flag & FLAG_ACK )
    {
        // Look for the message in pending_messages
        assert( pthread_mutex_lock( &pmg_mutex ) == 0 );
        for ( list<PendingMessage>::iterator i = pending_messages.begin(); i != pending_messages.end(); i ++ )
        {
            if ( ((const MsgHeader *)i->message)->id == msg_header->id )
            {
                // Remove the message from pending_messages
                delete [] i->message;
                pending_messages.erase( i );
                break;
            }
        }
        assert( pthread_mutex_unlock( &pmg_mutex ) == 0 );
    }
    else if ( msg_header->flag & FLAG_NEED_ACK )
    {
        // Send ack message
        MsgHeader ack;
        ack.signature = msg_signature;
        ack.nID = my_host_id;
        ack.type = msg_header->type;
        ack.flag = FLAG_ACK;
        ack.id = msg_header->id;
        ack.len = 0;
        assert( pthread_mutex_lock( &sr_mutex ) == 0 );
        if ( send( sock_report, &ack, sizeof( MsgHeader ), 0 ) < 0 )
            perror( "send" );
        assert( pthread_mutex_unlock( &sr_mutex ) == 0 );
    }
}


// Create a new message packet, setting the header and allocating buffer for payload

char * CreateMessage( short type, long id, int payload_len )
{
    char * message = new char[ sizeof( MsgHeader ) + payload_len ];
    if ( message )
    {
        MsgHeader * msg_header = (MsgHeader *)message;
        msg_header->signature = msg_signature;
        msg_header->nID = my_host_id;
        msg_header->type = type;
        msg_header->id = id;
        msg_header->flag = 0;
        msg_header->len = payload_len;
    }
    return message;
}


static int Timeout( short msg_type )
{
    switch ( msg_type )
    {
    default:
        return TIMEOUT;
    }
}


// Send a message

void SendMessage( char * message, bool need_ack )
{
    MsgHeader * msg_header = (MsgHeader *)message;
    // Debug
    //fprintf( stderr, "Send message, nID=%d, type=%d, id=%ld\n", msg_header->nID, msg_header->type, msg_header->id );
    if ( need_ack )
    {
        msg_header->flag |= FLAG_NEED_ACK;
        PendingMessage p_msg;
        p_msg.last_try = time( NULL );
        p_msg.times_tried = 1;
        p_msg.dID = 0;
        p_msg.timeout = Timeout( msg_header->type );
        p_msg.message = message;
        assert( pthread_mutex_lock( &pmg_mutex ) == 0 );
        pending_messages.push_back( p_msg );
        assert( pthread_mutex_unlock( &pmg_mutex ) == 0 );
    }
    assert( pthread_mutex_lock( &sr_mutex ) == 0 );
    if ( send( sock_report, message, sizeof( MsgHeader ) + ((const MsgHeader *)message)->len, 0 ) < 0 )
        perror( "send" );
    assert( pthread_mutex_unlock( &sr_mutex ) == 0 );
    if ( ! need_ack )
        delete [] message;
}


// Check the messages that have not yet been acknowledged. Resend a message if it is time out

void CheckPendingMessages( time_t current_time )
{
    assert( pthread_mutex_lock( &pmg_mutex ) == 0 );
    for ( list<PendingMessage>::iterator i = pending_messages.begin(); i != pending_messages.end(); )
    {
        if ( i->last_try <= current_time - i->timeout )
        {
            MsgHeader * msg_header = (MsgHeader *)((const MsgHeader *)i->message);
            msg_header->flag |= FLAG_RESEND;
            IF_DEBUG( 6, fprintf( stderr, "%s Resend message, type=%d, id=%ld\n", GetMeasurementTime(), msg_header->type, msg_header->id ) );
            assert( pthread_mutex_lock( &sr_mutex ) == 0 );
            if ( send( sock_report, i->message, sizeof( MsgHeader ) + msg_header->len, 0 ) < 0 )
                perror( "send" );
            assert( pthread_mutex_unlock( &sr_mutex ) == 0 );
            i->last_try = current_time;
            i->times_tried ++;
        }
        if ( i->times_tried >= MAX_RETRY )
        {
            delete [] i->message;
            i = pending_messages.erase( i );
        }
        else
            i ++;
    }
    assert( pthread_mutex_unlock( &pmg_mutex ) == 0 );
}


void ReportTest( void )
{
    /*
    for ( int i = 0; i < 1; i ++ )
    {
        char * message = CreateMessage( MSG_TMON, i, 16 );
        if ( message )
        {
            strncpy( message + sizeof( MsgHeader ), "1234567890123456", 16 );
            SendMessage( message );
        }
    }
    */
}


void ReportInitialize( void )
{
    message_id = rand();    // Set the initial message id
    pthread_mutex_init( &pmg_mutex, NULL );
    pthread_mutex_init( &sr_mutex, NULL );
    pthread_mutex_init( &mid_mutex, NULL );
    SetReportSocket();
}


void ReportFinalize( void )
{
    close( sock_report );
    pthread_mutex_destroy( &mid_mutex );
    pthread_mutex_destroy( &pmg_mutex );
    pthread_mutex_destroy( &sr_mutex );
}


long GetNextMessageID( void )
{
    assert( pthread_mutex_lock( &mid_mutex ) == 0 );
    long next_id = ++ message_id;
    assert( pthread_mutex_unlock( &mid_mutex ) == 0 );
    return next_id;
}
