#include "Pong.h"

int main( int argc, char **argv )
{
    ParseCommand( argc, argv );

    pong = new Pong( true );
    pong->InitializeSender();
    
    pong->SendProbes();

    pong->Finalize();

    return 0;
}


void Usage( char *argv0 )
{
    fprintf( stdout, "Usage: %s [-d debuglevel] [-l host] [-i interface] [-g] [-n] [-s] [-w dir] [-p port] [-b probe_num] [-u hop_num]\n", argv0 );
    fprintf( stdout, "  -d debuglevel: set initial debugging level\n" );
    fprintf( stdout, "        -l host: specify local ip/host\n" );
    fprintf( stdout, "   -i interface: specify local interface\n" );
    fprintf( stdout, "             -g: ignore probe configuration history\n" );
    fprintf( stdout, "         -w dir: set directory used to save measurement data, default is 'data/'\n" );
    fprintf( stdout, "             -n: disable interactive commands\n" );
    fprintf( stdout, "             -s: suppress clock skew information in debug output\n" );
    fprintf( stdout, "        -p port: set the control channel port of the sender side\n" );
    fprintf( stdout, "   -b probe_num: set mininum probe number of acceptable paths\n" );
    fprintf( stdout, "     -u hop_num: set mininum hop number of acceptable paths\n" );
    fprintf( stdout, "     -r host_id: set host_id of acceptable receiver\n" );
}


void ParseCommand( int argc, char **argv )
{
    int c;
    while ( ( c = getopt( argc, argv, "d:l:i:gw:nsp:b:u:r:" ) ) != -1 )
    {
	    switch ( c )
        {

       	// d: set debugging level
	    case 'd':
            {
	            DebugLevel = atoi( optarg );
	            break;
	        }

        case 'p':
            {
	            sender_control_port = atoi( optarg );
	            break;
	        }

	    case 'b':
            {
	            min_probe_num = atoi( optarg );
	            break;
	        }

	    case 'u':
            {
	            min_hop = atoi( optarg );
	            break;
	        }

        case 'r':
            {
	            receiver_id = atoi( optarg );
	            break;
	        }

        // l: set local origin of probes
	    case 'l':
            {
	            OriginHost = strdup( optarg );
	            if ( OriginHost == NULL )
                {
		            fprintf( stderr, "Couldn't allocate space for origin hostname\n" );
		            assert(0);
	            }
	            break;
	        }

        case 'g':
            {
                IgnoreConfiguration = true;
                break;
            }

        case 'n':
            {
                enable_interactive_command = false;
                break;
            }

        case 's':
            {
                suppress_clock_skew_info = true;
                break;
            }

        case 'i':
            {
#ifdef HAVE_PCAP
	            Interface = strdup( optarg );
	            if ( Interface == NULL )
                {
		            fprintf(stderr, "Couldn't allocate space for interface name\n");
		            assert(0);
	            }
	            break;
#else
	            fprintf( stderr, "pcap unavailable in this build" );
	            exit(1);
#endif // HAVE_PCAP
	        }

        case 'w':
            {
                mmt_data_dir = strdup( optarg );
                break;
            }

	    // Didn't know how to handle this case
	    default:
            {
                Usage( argv[0] );
	            exit(1);
	        }
        }
    }
}


void Pong::InitializeSender( void )
{
    SetProbeName( 'f', 'b', 's' );

    SetControlSocket_Snd();
    WaitForReceiver();

    EstimateRTT_Respond();
    e2e_rtt = RecvDoubleValue();
    IF_DEBUG( 3, fprintf( stderr, "Average round-trip time = %.2f(ms)\n", e2e_rtt / 1000 ) );

    Initialize_Common();
}

