#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <unistd.h>
//#include <sys/socket.h>
//#include <netinet/in.h>
#include <arpa/inet.h>
#include "ip_to_as.h"

double probe_period;
int probe_num, hop_num, hop_num_2;
HopInfo_with_AS hop_info[64], hop_info_2[64];
PrefixTree pt;

int main(int argc, char *argv[])
{
    char ch;
    
    if( argc <= 1 )
        usage();
    while ( ( ch = getopt( argc, argv, "h" ) ) != -1 )
    {
        switch(ch) {
            case 'h':
                usage();
                break;

            default:
                usage();
            break;
        }
    }

    if ( ! pt.readASFile( argv[1] ) )
        exit(1);

    const char * file = "history/summary_snd";
    FILE *fp;
    if ( ( fp = fopen( file, "r" ) ) == NULL )
    {
        fprintf( stderr, "Fail to read file '%s'\n", file );
        exit(1);
    }

    char buffer[1024], dir[200], path[200];
    char *sender, *st_time, *str_duration;
    int duration;
    while( ( fgets( buffer, 1024, fp ) ) != NULL  )
    {
        sender = strtok( buffer, "\t" );
        assert( sender != NULL );
        st_time = strtok( NULL, "\t" );
        assert( st_time != NULL );
        str_duration = strtok( NULL, "\t" );
        assert( str_duration != NULL );
        duration = atoi( str_duration );
        assert( duration >= 0 );
        //fprintf( stderr, "sender=%s, st_time=%s, duration=%d\n", sender, st_time, duration );
        strcpy( dir, "history/sender/" );
        strcat( dir, sender );
        strcat( dir, "/" );
        strcat( dir, st_time );
        fprintf( stderr, "- Process record: %s ... ", dir );
        strcpy( path, dir );
        strcat( path, "/probe_config" );
        if ( LoadProbeConfiguration( path ) )
        {
            strcpy( path, dir );
            strcat( path, "/link_as_info" );
            ResolveAS( path );
        }
        fprintf( stderr, "done\n" );
    }

    fclose( fp );

    return 0;
}


PrefixNode::PrefixNode()
{
    m_asnumber = -1;
    m_father_mask = -1;
    m_leaf = false;

    m_mask = 0;
    m_prefix = 0xffffffff;

    m_son_num = 0;
    m_son_num_max = 0;
    m_sons = NULL;
    
    m_string = NULL;
}

PrefixNode::PrefixNode(unsigned int ip, unsigned int mask, int as)
{
    PrefixNode();
    m_prefix = ip;
    m_mask = mask;
    m_asnumber = as;
    m_leaf = true;
}

PrefixNode::~PrefixNode()
{
    if(m_sons)
        delete [] m_sons;
    if(m_string)
        delete m_string;
}

void PrefixNode::destroy()
{
    if(m_son_num > 0) {
        for(int i=0; i<m_son_num; i++) {
            m_sons[i]->destroy();
            delete m_sons[i];
            m_sons[i] = NULL;
        }
    }
    if(m_sons)
        delete [] m_sons;
    m_sons = NULL;
}

int PrefixNode::bitmatch(unsigned ip1, unsigned ip2, int end, int start)
{
    int bit = 1 << (32 - start);
    int i = 0;
    for(i=start; i<=end; i++) {
        if( (ip1 & bit) != (ip2 & bit) )
            break;
        bit >>= 1;
    }    
    return i-1;
}

void PrefixNode::extendSonStorage()
{
    assert(m_son_num == m_son_num_max);
    m_son_num_max += SON_NUMBER_STEP;
    PrefixNode **tmp = new PrefixNode*[m_son_num_max];
    memcpy(tmp, m_sons, sizeof(PrefixNode*) * m_son_num);
    delete m_sons;
    m_sons = tmp;
}

void PrefixNode::addNode(PrefixNode *pn) 
{
    int i, j, max_match, match_id;
    // check match
    if(m_mask != 0) {     // only root has m_mask as 0
        assert(pn->m_mask >= m_mask);
        i = MASK_NUMBER(m_mask);
        assert( (pn->m_prefix & i) == (m_prefix & i) );
    }

    if(m_son_num == 0) {
        m_son_num = 1;
        m_son_num_max = SON_NUMBER_STEP;
        m_sons = new  PrefixNode*[SON_NUMBER_STEP];
        m_sons[0] = pn;
        pn->m_leaf = true;
        return;
    }
    
    max_match = 0;
    match_id = -1;
    for(i=0; i<m_son_num; i++) {
    //    fprintf(stderr, "compare: pn->m_string: %s, m_sons[i]->m_string:%s\n", pn->m_string, m_sons[i]->m_string);
        j = bitmatch(pn->m_prefix, m_sons[i]->m_prefix, MIN(pn->m_mask, m_sons[i]->m_mask), 1);        
        if(j > max_match) {
            max_match = j;
            match_id = i;
        }
    //    fprintf(stderr, "match: i:%d, j:%d, max_match:%d, match_id:%d, min:%d\n\n", i, j, max_match, match_id, MIN(pn->m_mask, m_sons[i]->m_mask));
    }
    if(max_match < 4)
        max_match = 0;
    assert(max_match >= m_mask);
    if(max_match == m_mask) {    // add a new son
        if(m_son_num == m_son_num_max) {    // if storage used out, extend it
            extendSonStorage();
        }
        m_sons[m_son_num++] = pn;
    }
    else {    
        PrefixNode *match_node = m_sons[match_id];
        int min_mask = MIN(pn->m_mask, m_sons[match_id]->m_mask);
        
        if(max_match > min_mask)
            fprintf(stderr, "max_match:%d, min_mask:%d\n", max_match, min_mask);
        assert(max_match <= min_mask);
        if(max_match == min_mask) {    
            if(pn->m_mask == m_sons[match_id]->m_mask) {    // equal prefix and mask, should be the same entry
                if(m_sons[match_id]->m_leaf) {
                    if( m_sons[match_id]->m_asnumber == -1 ) {
                        m_sons[match_id]->m_asnumber = pn->m_asnumber;
                        m_sons[match_id]->m_leaf = true;
                    }
                    else if( m_sons[match_id]->m_asnumber != pn->m_asnumber) {
                        /*
                        struct in_addr ip;
                        ip.s_addr = htonl( m_sons[match_id]->m_prefix );
                        fprintf(stderr, "Inconsistent entry, same prefix but different as, prefix: %s, as1: %d, as2:%d\n", inet_ntoa( ip ), m_sons[match_id]->m_asnumber, pn->m_asnumber);
                        */
                        if ( m_sons[match_id]->m_asnumber < pn->m_asnumber )
                            m_sons[match_id]->m_asnumber = pn->m_asnumber;
                    }
                    else {        // consistent, so return;
                    //    fprintf(stderr, "no new node added because the same\n");
                        return;
                    }
                }
                else {
                    m_sons[match_id]->m_leaf = true;
                //    fprintf(stderr, "change a non-leaf to leaf\n");
                    return;
                }
            }
            else if(pn->m_mask < m_sons[match_id]->m_mask) {
                m_sons[match_id] = pn;
            //    fprintf(stderr, "pn become m_sons[match_id]'s father: pn->m_string: %s, m_sons[match_id]->m_string:%s\n", pn->m_string, m_sons[match_id]->m_string);
                pn->addNode(match_node);                
                return;
            }
            else {    // pn->m_mask > m_sons[match_id]->m_mask                
            //    fprintf(stderr, "pn become m_sons[match_id]'s son: pn->m_string: %s, m_sons[match_id]->m_string:%s\n", pn->m_string, m_sons[match_id]->m_string);
                m_sons[match_id]->addNode(pn);
                return;
            }
        }
        else {    // split my son
            char tmpstr[64];
            PrefixNode *newson = new PrefixNode();
            newson->m_prefix = pn->m_prefix;
            newson->m_mask = max_match;                        
            newson->m_string = new char[64];
            assert(pn->m_string && strlen(pn->m_string));
            
            strcpy(tmpstr, pn->m_string);
            for(i=0; tmpstr[i] != 0; i++) {
                if(tmpstr[i] == '/')
                    tmpstr[i] = 0;
            }
            sprintf(newson->m_string, "V %s/%d -1\n", tmpstr, max_match);
            
        //    fprintf(stderr, "split: pn->m_string: %s, m_sons[match_id]->m_string:%s, newson->m_string:%s\n", pn->m_string, m_sons[match_id]->m_string, newson->m_string);
            m_sons[match_id] = newson;
            newson->addNode(match_node);            
            newson->addNode(pn);
            return;
        }
        
    }
}

int PrefixNode::IptoAS(unsigned int ip)
{
    int nmask, as;
    char *nstring;
    assert(m_mask == 0);
    as = IptoAS(ip, nmask, nstring);
    /*if( as != -1 )
        fprintf(stderr, "%s ", nstring);*/
    return as;
}

int PrefixNode::IptoAS(unsigned int ip, int &nmask, char *&nstring)
{
    int i, max_match, match_id, match_as;
    int son_as_id, my_nmask;
    char *pstring = NULL;
    
    // check match
    if(m_mask != 0) {     // only root has m_mask as 0
        i = MASK_NUMBER(m_mask);
        if((ip & i) != (m_prefix & i))  {    // unmatched node
            nmask = 0;
            return -1;            
        }
    }
    
    max_match = 0;
    match_id = -1;
    my_nmask = 0;
    match_as = -1;
    for(i=0; i<m_son_num; i++) {
    //    fprintf(stderr, "compare: pn->m_string: %s, m_sons[i]->m_string:%s\n", pn->m_string, m_sons[i]->m_string);
    //    fprintf(stderr, "%s call %s\n", m_string, m_sons[i]->m_string);
        son_as_id = m_sons[i]->IptoAS(ip, my_nmask, pstring);
        if(son_as_id != -1) {
            if(my_nmask > max_match) {                
                max_match = my_nmask;
                match_as = son_as_id;
                nstring = pstring;
            }
        }
        
    //    fprintf(stderr, "match: i:%d, j:%d, max_match:%d, match_id:%d, min:%d\n\n", i, j, max_match, match_id, MIN(pn->m_mask, m_sons[i]->m_mask));
    }

//    fprintf(stderr, "\n\nm_string:%s, max_match: %d, match_as: %d\n", m_string, max_match, match_as);
    
    if(match_as == -1) {    // no son has longer match
        if(m_leaf && m_asnumber != -1) {
            nmask = m_mask;
            nstring = m_string;
        //    fprintf(stderr, "\nMy self: %s %d %d\n", m_string, m_asnumber, m_mask);
            return m_asnumber;
        }
        else
            return -1;
    }
    else {
    //    fprintf(stderr, "\nFrom son: %d %d\n", match_as, max_match);
        nmask = max_match;
        return match_as;
    }
}

PrefixTree::PrefixTree()
{
    m_root = NULL;
}

PrefixTree::~PrefixTree()
{
    if(m_root)
        m_root->destroy();
    delete m_root;
}

bool PrefixTree::readASFile(const char *fname)
{
    char line[256];
    int counter;
    int i, j, k;
    char *str_ip, *str_mask, *str_as, *str;
    unsigned int ip, mask;
    int as;
    FILE *f = fopen(fname, "r");
    bool ret = true;
    if(f == NULL) {
        fprintf(stderr, "Open file '%s' error: %s\n", fname, strerror(errno));
        return false;
    }
    else {
        fprintf(stderr, "Open AS file '%s' success\n", fname);
    }

    if(m_root)
        m_root->destroy();
    delete m_root;
    m_root = new PrefixNode();

    counter = 0;
    while( !feof(f) ) {
    //    assert(counter ++ < 4);
        if( !fgets(line, 255, f) ) {
            if(feof(f)) 
                break;
            else {                
                fprintf(stderr, "Read file %s error: %s\n", fname, strerror(errno));
                ret = false;
                break;
            }
        }        
        
        //64.156.200.0/21 3356
        i = strlen(line);
        for(j=0; j<i; j++)
            if(line[j] == '/')
                break;
        for(k=j+1; k<i; k++)
            if(line[k] == ' ' || line[k] == '\t')
                break;
        if(j>=i || k>=i) {
            fprintf(stderr, "This line has wrong format: %s, i:%d, j:%d, k:%d\n", line, i, j, k);
            ret = false;
            break;    
        }
        
        str = new char[strlen(line) + 2];
        strcpy(str, line);
        
        str_ip = line;
        line[j] = '\0';
        str_mask = line + j + 1;
        line[k] = '\0';
        str_as = line + k + 1;

        ip = net_addr(str_ip);
        mask = atoi(str_mask);
        as = atoi(str_as);
        
    //    fprintf(stderr, "%s\n", str);
    //    fprintf(stderr, "%x %d %d\n", ip, mask, as);    
        PrefixNode *pn = new PrefixNode(ip, mask, as);
        pn->m_string = str;
        addNode(pn);
    }
    fclose(f);
    return ret;
}

void PrefixTree::addNode(PrefixNode *pn)
{
    m_root->addNode(pn);
}

int PrefixTree::net_addr(const char *saddr)
{
    int iaddr = 0;
    int step = 24;
    int i,j;
    int len = strlen(saddr);
    char *s = new char[len+2], *sip;
    strcpy(s, saddr);
    sip = s;
    for(i=0; i<len+1; i++) {
        if(s[i] == '.' || s[i] == '\0') {
            assert(step >= 0);
            s[i] = '\0';
            j = atoi(sip);
            iaddr |= (j<<step);
            sip = s + i + 1;
            step -=8; 
        }
    }
    delete s;
    return iaddr;
}

int PrefixTree::IptoAS(unsigned int ip)
{
    return m_root->IptoAS(ip);
}

int PrefixTree::IptoAS(const char *str_ip)
{
    int ip = net_addr(str_ip);
    return IptoAS(ip);
}

void usage()
{
    fprintf(stderr, "resolve_as <map_file>\n");
    exit(1);
}


// Load probe configuration from file

bool LoadProbeConfiguration( const char * probe_config_file )
{
    FILE * fp;
    if( ( fp = fopen( probe_config_file, "r" ) ) == NULL )
    {
        fprintf( stderr, "Cannot open configuration file '%s'\n", probe_config_file );
        return false;
    }

    if ( ReadParameter( fp, "probe_period", "%lf", &probe_period ) != 0 )
    {
        fprintf( stderr, "Invalid configuration (read 'probe_period')\n" );
        exit(1);
    }
    if ( ReadParameter( fp, "probe_num", "%d", &probe_num ) != 0 )
    {
        fprintf( stderr, "Invalid configuration (read 'probe_num')\n" );
        exit(1);
    }

    char buffer[MAX_CONFIG_LINE_WIDTH + 1];
    fgets( buffer, MAX_CONFIG_LINE_WIDTH, fp );

    for ( int i = 0; i < probe_num; i ++ )
    {
        if ( ! fgets( buffer, MAX_CONFIG_LINE_WIDTH, fp ) )
        {
            fprintf( stderr, "Invalid configuration\n" );
            exit(1);
        }
    }

    HopInfo_with_AS *hi;
    char ip[20], host_name[300];

    if ( ReadParameter( fp, "hop_num", "%d", &hop_num ) != 0 )
    {
        fprintf( stderr, "Invalid configuration (read 'hop_num')\n" );
        exit(1);
    }
    
    fgets( buffer, MAX_CONFIG_LINE_WIDTH, fp );
    
    for ( int i=0; i < hop_num; i ++ )
    {
        if ( ! fgets( buffer, MAX_CONFIG_LINE_WIDTH, fp ) )
        {
            fprintf( stderr, "Invalid configuration\n" );
            exit(1);
        }
        hi = hop_info + i;
        sscanf( buffer, "%d %s %s", &(hi->ttl), ip, host_name );
        hi->ip = strdup( ip );
        hi->host_name = strdup( host_name );
    }

    if ( ReadParameter( fp, "hop_num_2", "%d", &hop_num_2 ) != 0 )
    {
        fprintf( stderr, "Invalid configuration (read 'hop_num_2')\n" );
        exit(1);
    }

    fgets( buffer, MAX_CONFIG_LINE_WIDTH, fp );

    for ( int i=0; i < hop_num_2; i ++ )
    {
        if ( ! fgets( buffer, MAX_CONFIG_LINE_WIDTH, fp ) )
        {
            fprintf( stderr, "Invalid configuration\n" );
            exit(1);
        }
        hi = hop_info_2 + i;
        sscanf( buffer, "%d %s %s", &(hi->ttl), ip, host_name );
        hi->ip = strdup( ip );
        hi->host_name = strdup( host_name );
    }

    fclose( fp );
    return true;
}


// Read the next parameter in the configuration file

char ReadParameter( FILE * fp, const char * parameter_name, const char * format_string, void * parameter_ptr )
{
	char buffer[MAX_CONFIG_LINE_WIDTH + 1];
	char match = 0;
	while ( ! match && fgets( buffer, MAX_CONFIG_LINE_WIDTH, fp ) )
	{
		if ( buffer[0] != '#' ) // Not a comment line
		{
			int str_len = strlen( parameter_name );
			if ( strncmp( buffer, parameter_name, str_len ) == 0 )
			{
				match = 1;
				sscanf( buffer + str_len + 1, format_string, parameter_ptr );
			}
		}
	}
	if ( ! match )
		return -1;
	return 0;
}


#define TYPE_UNKNOWN    0
#define INTRA_AS        1
#define INTER_AS        2

bool ResolveAS( const char * output_file )
{
    int asn[65], link_type[64], redundant_link[64];
    assert( hop_num > 1 && hop_num <= 63 );
    for ( int i = 0; i <= hop_num; i ++ )
        link_type[i] = TYPE_UNKNOWN;
    int resolved = 0;
    asn[0] = -1;
    redundant_link[0] = false;

    // Resolve ASN
    for ( int i = 0; i < hop_num; i ++ )
    {
        assert( hop_info[i].ttl == i + 1 );
        if ( strcmp( hop_info[i].ip, "0.0.0.0" ) != 0 )
        {
            redundant_link[i + 1] = false;
            asn[i + 1] = pt.IptoAS( hop_info[i].ip );
            if ( asn[i + 1] != -1 )
                resolved ++;
        }
        else
        {
            redundant_link[i + 1] = true;
            asn[i + 1] = -1;
        }
    }
    asn[0] = asn[1];
    asn[hop_num + 1] = asn[hop_num];

    // Figure out unresolved ASN if possible
    for ( int i = 0; i < hop_num; i ++ )
    {
        if ( asn[i + 1] == -1 )
        {
            int j, k;
            for ( j = i; j >= 0; j -- )
                if ( asn[j] != -1 )
                    break;
            for ( k = i + 2; k <= hop_num + 1; k ++ )
                if ( asn[k] != -1 )
                    break;
            if ( asn[j] == asn[k] )
            {
                for ( int l = j + 1; l < k; l ++ )
                    asn[l] = asn[j];
            }
        }
    }
 
    // Figure out link type
    for ( int i = 1; i < hop_num; i ++ )
    {
        if ( asn[i] != -1 && asn[i + 1] != -1 )
        {
            if ( asn[i] != asn[i + 1] )
                link_type[i - 1] = link_type[i] = INTER_AS;
            else
                link_type[i] = INTRA_AS;
        }
    }
    link_type[0] = link_type[hop_num] = INTRA_AS;

    // Output result
    FILE * fp;
    if( ( fp = fopen( output_file, "w" ) ) == NULL )
    {
        fprintf( stderr, "Cannot open file '%s' for writing\n", output_file );
        exit(1);
    }

    fprintf( fp, "probe_num=%d\thop_num=%d\thop_num_2=%d\tresolved=%d\n", probe_num, hop_num, hop_num_2, resolved );
    fprintf( fp, "asn:\n" );
    for ( int i = 0; i <= hop_num + 1; i ++ )
        fprintf( fp, "%d\t", asn[i] );
    fprintf( fp, "\n" );
    fprintf( fp, "link_type:\n" );
    for ( int i = 0; i <= hop_num; i ++ )
    {
        fprintf( fp, "%d\t", link_type[i] );
    }
    fprintf( fp, "\n" );
    fprintf( fp, "redundant_link:\n" );
    for ( int i = 0; i <= hop_num; i ++ )
    {
        fprintf( fp, "%d\t", redundant_link[i] );
    }
    fprintf( fp, "\n" );

    fclose( fp );
    return true;
}
